"""
"""

import sys

from .config import Config
from .gpu.decorator import GPU
from .gpu.torch import disable_cuda_intercept
from .gradio import gradio_auto_wrap
from .gradio import disable_gradio_auto_wrap
from .gradio import enable_gradio_auto_wrap


if Config.zero_gpu and sys.version_info.minor < 9: # pragma: no cover
    raise RuntimeError("Actually using @spaces.GPU on a ZeroGPU Space requires Python 3.9+")


__all__ = [
    'GPU',
    'disable_cuda_intercept',
    'gradio_auto_wrap',
    'disable_gradio_auto_wrap',
    'enable_gradio_auto_wrap',
]



"""
"""

from ..config import Config
from . import torch

if Config.zero_gpu:
    if torch.is_in_bad_fork():
        raise RuntimeError(
            "CUDA has been initialized before importing the `spaces` package"
        )
    torch.patch() # pragma: no cover



"""
"""
from __future__ import annotations

import time
import warnings
from dataclasses import asdict
from dataclasses import dataclass
from datetime import timedelta
from http import HTTPStatus

import gradio as gr
import requests

from .. import utils
from ..config import Config


CGROUP_PATH = utils.self_cgroup_device_path()
TOKEN_HEADER = 'X-IP-Token'

UNUSED_MESSAGE = "GPU device not used"


@dataclass
class ScheduleParams:
    cgroupPath: str
    taskId: int
    token: str
    durationSeconds: int | None

@dataclass
class ScheduleResponse:
    idle: bool
    nvidiaIndex: int
    nvidiaUUID: str

@dataclass
class ReleaseParams:
    cgroupPath: str
    taskId: int
    nvidiaIndex: int
    fail: bool


def base_url() -> str:
    assert Config.zero_device_api_url is not None
    return Config.zero_device_api_url

def post(path: str, params: dict | None = None) -> requests.Response:
    return requests.post(base_url() + path, params=params if params else None)


def startup_report():
    retries, max_retries = 0, 2
    while (status := post('/startup-report').status_code) == HTTPStatus.NOT_FOUND: # pragma: no cover
        time.sleep(1)
        if (retries := retries + 1) > max_retries:
            raise RuntimeError("Error while initializing ZeroGPU: NotFound")
    if status != HTTPStatus.OK: # pragma: no cover
        raise RuntimeError("Error while initializing ZeroGPU: Unknown")


def schedule(
    task_id: int,
    request: gr.Request,
    duration: timedelta | None,
) -> ScheduleResponse:

    headers = getattr(request, 'headers', None)
    if headers is None or not hasattr(headers, '__dict__'):
        raise gr.Error("Internal Gradio error")

    # Compatibility trick
    if not hasattr(headers, 'get'):
        headers = headers.__dict__ # pragma: no cover

    if not (token := headers.get(TOKEN_HEADER.lower())):
        raise gr.Error("Internal infra error")

    res = post('/schedule', params=asdict(ScheduleParams(
        cgroupPath=CGROUP_PATH,
        taskId=task_id,
        token=token,
        durationSeconds=duration.seconds if duration is not None else None,
    )))

    if res.status_code == HTTPStatus.TOO_MANY_REQUESTS:
        raise gr.Error("No GPU is currently available")

    try:
        data = res.json()
    except requests.JSONDecodeError: # pragma: no cover
        data = {}

    if not res.ok: # pragma: no cover
        raise RuntimeError(f"ZeroGPU API /schedule error: {data.get('detail')}")

    return ScheduleResponse(**data)


def release(
    task_id: int,
    nvidia_index: int,
    fail: bool = False,
) -> None:

    res = post('/release', params=asdict(ReleaseParams(
        cgroupPath=CGROUP_PATH,
        taskId=task_id,
        nvidiaIndex=nvidia_index,
        fail=fail,
    )))

    if res.status_code == HTTPStatus.NO_CONTENT: # pragma: no cover
        try:
            gr.Warning(UNUSED_MESSAGE)
        except AttributeError:
            pass
        warnings.warn(UNUSED_MESSAGE, RuntimeWarning)
        return None

    if not res.ok:
        try:
            data = res.json()
        except requests.JSONDecodeError: # pragma: no cover
            data = {}
        # TODO: use future gr.Warning() if /release didn't detect GPU usage
        raise RuntimeError(f"ZeroGPU API /release error: {data.get('detail')}")

    return None



"""
"""
from __future__ import annotations

import os

from .utils import boolean


class Settings:
    def __init__(self):
        self.zero_gpu = boolean(
            os.getenv('SPACES_ZERO_GPU'))
        self.zero_device_api_url = (
            os.getenv('SPACES_ZERO_DEVICE_API_URL'))
        self.gradio_auto_wrap = boolean(
            os.getenv('SPACES_GRADIO_AUTO_WRAP'))


Config = Settings()


if Config.zero_gpu:
    assert Config.zero_device_api_url is not None, (
        'SPACES_ZERO_DEVICE_API_URL env must be set '
        'on ZeroGPU Spaces (identified by SPACES_ZERO_GPU=true)'
    )



"""
"""
# pyright: reportPrivateImportUsage=false

from __future__ import annotations

import multiprocessing
import os
from concurrent.futures import ProcessPoolExecutor
from typing import TYPE_CHECKING
from typing import Tuple

if TYPE_CHECKING:
    import torch as Torch

try:
    import torch
except ImportError:
    torch = None


if torch is not None:

    _tensor_to      = torch.Tensor.to
    _tensor_cuda    = torch.Tensor.cuda
    _cuda_init      = torch._C._cuda_init
    _cuda_available = torch.cuda.is_available

    TensorToArgs = Tuple[torch.device, torch.dtype, bool, torch.memory_format]

    to_ops: list[tuple[Torch.Tensor, TensorToArgs]] = []
    cuda_ops: list[Torch.Tensor] = []

    def _to_op_register(self: Torch.Tensor, *args, **kwargs):
        parsed = torch._C._nn._parse_to(*args, **kwargs)
        device, *_ = parsed
        if not isinstance(device, torch.device):
            return _tensor_to(self, *args, **kwargs)
        if device.type != 'cuda':
            return _tensor_to(self, *args, **kwargs)
        to_ops.append((self, parsed))
        return self

    def _cuda_op_arg_check(device: Torch.device | int | str | None) -> bool:
        if device is None:
            return True
        if isinstance(device, int):
            return True
        if isinstance(device, str):
            device = torch.device(device)
        return device.type == 'cuda'

    def _cuda_op_register(self: Torch.Tensor, device: Torch.device | int | str | None = None, **kwargs):
        if not _cuda_op_arg_check(device):
            # Let PyTorch handle the fail
            return _tensor_cuda(self, device, **kwargs)
        cuda_ops.append(self)
        return self

    def _cuda_init_raise():
        raise RuntimeError(
            "CUDA must not be initialized in the main process "
            "on Spaces with Stateless GPU environment.\n"
            "You can look at this Stacktrace to find out "
            "which part of your code triggered a CUDA init"
        )

    def _patch():
        torch.Tensor.to         = _to_op_register   # type: ignore
        torch.Tensor.cuda       = _cuda_op_register # type: ignore
        torch._C._cuda_init     = _cuda_init_raise
        torch.cuda.is_available = lambda: True

    def _unpatch():
        torch.Tensor.to         = _tensor_to
        torch.Tensor.cuda       = _tensor_cuda
        torch._C._cuda_init     = _cuda_init
        torch.cuda.is_available = _cuda_available

    def _move(nvidia_uuid: str):
        os.environ['CUDA_VISIBLE_DEVICES'] = nvidia_uuid
        for op in to_ops:
            tensor, parsed_args = op
            _, dtype, _, memory_format = parsed_args
            tensor.data = _tensor_to(tensor,
                device='cuda',
                dtype=dtype,
                memory_format=memory_format,
            ) # type: ignore
        for op in cuda_ops:
            tensor = op
            tensor.data = _tensor_cuda(tensor)
        torch.cuda.init()

    def _is_in_bad_fork():
        with ProcessPoolExecutor(mp_context=multiprocessing.get_context('fork')) as e:
            f = e.submit(torch.cuda._is_in_bad_fork)
            return f.result()

    def _disable_cuda_intercept():
        torch.Tensor.to   = _tensor_to
        torch.Tensor.cuda = _tensor_cuda

else:

    _patch = lambda: None
    _unpatch = lambda: None
    _move = lambda nvidia_uuid: None
    _is_in_bad_fork = lambda: False
    _disable_cuda_intercept = lambda: None


patch = _patch
unpatch = _unpatch
move = _move
is_in_bad_fork = _is_in_bad_fork
disable_cuda_intercept = _disable_cuda_intercept



"""
"""
from __future__ import annotations

import multiprocessing
from multiprocessing.queues import SimpleQueue as _SimpleQueue
from pathlib import Path
from pickle import PicklingError
from typing import Callable
from typing import TypeVar


T = TypeVar('T')


def self_cgroup_device_path() -> str:
    cgroup_content = Path('/proc/self/cgroup').read_text()
    for line in cgroup_content.strip().split('\n'):
        contents = line.split(':devices:')
        if len(contents) != 2:
            continue # pragma: no cover
        return contents[1]
    raise Exception # pragma: no cover


class SimpleQueue(_SimpleQueue[T]):
    def __init__(self, *args):
        super().__init__(*args, ctx=multiprocessing.get_context('fork'))
    def put(self, element: T):
        try:
            super().put(element)
        except PicklingError:
            raise # pragma: no cover
        # https://bugs.python.org/issue29187
        except Exception as e:
            message = str(e)
            if not "pickle" in message:
                raise # pragma: no cover
            raise PicklingError(message)
    def close(self): # Python 3.8 static typing trick
        super().close() # type: ignore


def drop_params(fn: Callable[[], T]) -> Callable[..., T]:
    def drop(*args):
        return fn()
    return drop


def boolean(value: str | None) -> bool:
    return value is not None and value.lower() in ("1", "t", "true")



"""
"""
from __future__ import annotations

import multiprocessing
import inspect
import os
import signal
import traceback
import uuid
from concurrent.futures import ThreadPoolExecutor
from datetime import timedelta
from functools import partial
from inspect import Parameter
from pickle import PicklingError
from threading import Thread
from typing import TYPE_CHECKING
from typing import Any
from typing import Dict
from typing import Callable
from typing import Generator
from typing import Generic
from typing import Tuple
from typing import TypeVar
from typing_extensions import Concatenate
from typing_extensions import ParamSpec

import gradio as gr
import psutil

from ..utils import drop_params
from ..utils import SimpleQueue as Queue
from . import client
from . import torch


Process = multiprocessing.get_context('fork').Process

Arg = Tuple[Tuple[Any], Dict[str, Any]]
Res = TypeVar('Res')
Param = ParamSpec('Param')
NvidiaIndex = int
NvidiaUUID = str


class Worker(Generic[Res]):
    process: Process
    arg_queue: Queue[Arg]
    res_queue: Queue[Res]
    _sentinel: Thread

    def __init__(
        self,
        target: Callable[[Queue[Arg], Queue[Res], NvidiaUUID, list[int]], None],
        nvidia_uuid: str,
    ):
        self._sentinel = Thread(target=self._close_on_exit)
        self.arg_queue = Queue()
        self.res_queue = Queue()
        fds = [c.fd for c in psutil.Process().connections()]
        args = self.arg_queue, self.res_queue, nvidia_uuid, fds
        if TYPE_CHECKING:
            target(*args)
        self.process = Process(
            target=target,
            args=args,
            daemon=True,
        )
        self.process.start()
        self._sentinel.start()

    def _close_on_exit(self):
        self.process.join()
        self.res_queue.close()


def regular_function_wrapper(
    task: Callable[Param, Res],
    duration: timedelta | None,
) -> Callable[Concatenate[gr.Request, Param], Res]:

    workers: dict[NvidiaIndex, Worker[list[Res] | Exception]] = {}
    task_id = id(task)

    def gradio_handler(request: gr.Request, /, *args: Param.args, **kwargs: Param.kwargs) -> Res:

        schedule_response = client.schedule(task_id, request, duration)
        nvidia_index = schedule_response.nvidiaIndex
        nvidia_uuid = schedule_response.nvidiaUUID
        release = partial(client.release, task_id=task_id, nvidia_index=nvidia_index)

        worker = workers.get(nvidia_index)
        if worker is None or not worker.process.is_alive():
            worker = Worker(thread_wrapper, nvidia_uuid)
            workers[nvidia_index] = worker

        try:
            worker.arg_queue.put((args, kwargs))
        except PicklingError:
            release(fail=True)
            raise

        try:
            res = worker.res_queue.get()
        except EOFError:
            release(fail=True)
            raise gr.Error("GPU task aborted")
        if isinstance(res, Exception):
            release(fail=True)
            raise res
        release()
        return res[0]


    def thread_wrapper(
        arg_queue: Queue[Arg],
        res_queue: Queue[list[Res] | Exception],
        nvidia_uuid: str,
        fds: list[int],
    ):
        torch.unpatch()
        torch.move(nvidia_uuid)
        for fd in fds:
            os.close(fd)
        signal.signal(signal.SIGTERM, drop_params(arg_queue.close))
        while True:
            try:
                args, kwargs = arg_queue.get()
            except OSError:
                break
            with ThreadPoolExecutor() as executor:
                future = executor.submit(task, *args, **kwargs) # type: ignore
            try:
                res = [future.result()]
            except Exception as e:
                traceback.print_exc()
                res = e
            try:
                res_queue.put(res)
            except PicklingError as e:
                res_queue.put(e)
    
    update_handler_meta(gradio_handler, inspect.signature(task))

    return gradio_handler


def generator_function_wrapper(
    task: Callable[Param, Generator[Res, None, None]],
    duration: timedelta | None,
) -> Callable[Concatenate[gr.Request, Param], Generator[Res, None, None]]:

    workers: dict[NvidiaIndex, Worker[list[Res] | Exception | None]] = {}
    task_id = id(task)

    def gradio_handler(request: gr.Request, /, *args: Param.args, **kwargs: Param.kwargs) -> Generator[Res, None, None]:

        schedule_response = client.schedule(task_id, request, duration)
        nvidia_index = schedule_response.nvidiaIndex
        nvidia_uuid = schedule_response.nvidiaUUID
        release = partial(client.release, task_id=task_id, nvidia_index=nvidia_index)

        worker = workers.get(nvidia_index)
        if worker is None or not worker.process.is_alive():
            worker = Worker(thread_wrapper, nvidia_uuid)
            workers[nvidia_index] = worker

        try:
            worker.arg_queue.put((args, kwargs))
        except PicklingError:
            release(fail=True)
            raise

        yield_queue: Queue[list[Res] | Exception | None] = Queue()
        def fill_yield_queue(worker: Worker[list[Res] | Exception | None]):
            while True:
                try:
                    res = worker.res_queue.get()
                except Exception:
                    release(fail=True)
                    yield_queue.close()
                    return
                if isinstance(res, Exception):
                    release(fail=True)
                    yield_queue.put(res)
                    return
                if res is None:
                    release()
                    yield_queue.put(None)
                    return
                yield_queue.put(res)

        with ThreadPoolExecutor() as e:
            e.submit(fill_yield_queue, worker)
            while True:
                try:
                    res = yield_queue.get()
                except Exception:
                    raise gr.Error("GPU task aborted")
                if isinstance(res, Exception):
                    raise res
                if res is None:
                    break
                yield res[0]


    def thread_wrapper(
        arg_queue: Queue[Arg],
        res_queue: Queue[list[Res] | Exception | None],
        nvidia_uuid: str,
        fds: list[int],
    ):
        torch.unpatch()
        torch.move(nvidia_uuid)
        for fd in fds:
            os.close(fd)
        signal.signal(signal.SIGTERM, drop_params(arg_queue.close))
        while True:
            try:
                args, kwargs = arg_queue.get()
            except OSError:
                break
            def iterate():
                gen = task(*args, **kwargs) # type: ignore
                while True:
                    try:
                        res = next(gen)
                    except StopIteration:
                        break
                    except Exception as e:
                        res_queue.put(e)
                        break
                    try:
                        res_queue.put([res])
                    except PicklingError as e:
                        res_queue.put(e)
                        break
                    else:
                        continue
            with ThreadPoolExecutor() as executor:
                executor.submit(iterate)
            res_queue.put(None)

    update_handler_meta(gradio_handler, inspect.signature(task))

    return gradio_handler


def update_handler_meta(handler: Callable, task_signature: inspect.Signature):
    param_name = f'request_{uuid.uuid4().hex}'
    new_param = Parameter(param_name, Parameter.POSITIONAL_ONLY, annotation=gr.Request)
    new_params = [new_param] + list(task_signature.parameters.values())
    handler.__signature__ = task_signature.replace(parameters=new_params)
    handler.__annotations__.update({param_name: gr.Request})



"""
"""
from __future__ import annotations

from typing import Callable
from typing import Generator
from typing import TypeVar
from typing import overload
from typing_extensions import ParamSpec

from .config import Config
from .gpu.decorator import GPU


Param = ParamSpec('Param')
Res = TypeVar('Res')


gradio_auto_wrap_enabled = Config.gradio_auto_wrap


def disable_gradio_auto_wrap():
    global gradio_auto_wrap_enabled
    gradio_auto_wrap_enabled = False

def enable_gradio_auto_wrap():
    global gradio_auto_wrap_enabled
    gradio_auto_wrap_enabled = True


@overload
def gradio_auto_wrap(
    task:
     Callable[Param, Res],
) -> Callable[Param, Res]:
    ...
@overload
def gradio_auto_wrap(
    task:
     None,
) -> None:
    ...
def gradio_auto_wrap(
    task:
      Callable[Param, Res]
    | None,
) -> (Callable[Param, Res]
    | None):
    """
    """
    if not gradio_auto_wrap_enabled:
        return task
    if not callable(task):
        return task
    return GPU(task) # type: ignore



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['spaces', 'spaces.gpu']

package_data = \
{'': ['*']}

install_requires = \
['gradio>=3.2,<4.0',
 'psutil>=2,<6',
 'requests>=2.19,<3.0',
 'typing-extensions>=4,<5']

setup_kwargs = {
    'name': 'spaces',
    'version': '0.12b1',
    'description': 'Utilities for Hugging Face Spaces',
    'long_description': '# Hugging Face Spaces\n\n## Installation\n\n`pip install spaces`\n',
    'author': 'Charles Bensimon',
    'author_email': 'charles@huggingface.co',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://huggingface.co',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
"""
from __future__ import annotations

import inspect
from datetime import timedelta
from functools import partial
from typing import Callable
from typing import TypeVar
from typing import overload
from typing_extensions import Concatenate
from typing_extensions import ParamSpec

import gradio as gr

from ..config import Config
from . import client
from .wrappers import regular_function_wrapper
from .wrappers import generator_function_wrapper


P = ParamSpec('P')
R = TypeVar('R')


decorated_cache: dict[Callable, Callable] = {}


@overload
def GPU(
    task: None = None, *,
    duration: int | timedelta | None = None,
) -> Callable[[Callable[P, R]], Callable[Concatenate[gr.Request, P], R]]:
    ...
@overload
def GPU(
    task: Callable[P, R], *,
    duration: int | timedelta | None = None,
) -> Callable[Concatenate[gr.Request, P], R]:
    ...
def GPU(
    task: Callable[P, R] | None = None, *,
    duration: int | timedelta | None = None,
) -> Callable[[Callable[P, R]], Callable[Concatenate[gr.Request, P], R]] | Callable[Concatenate[gr.Request, P], R]:
    """
    ZeroGPU decorator

    Basic usage:
        ```
        @spaces.GPU
        def fn(...):
            # CUDA is available here
            pass
        ```

    With custom duration:
        ```
        @spaces.GPU(duration=45) # Expressed in seconds
        def fn(...):
            # CUDA is available here
            pass
        ```

    Args:
        task (`Callable | None`): Python function that requires CUDA
        duration (`int | datetime.timedelta`): Estimated duration in seconds or `datetime.timedelta`

    Returns:
        `Callable`: GPU-ready function
    """
    if task is None:
        return partial(_GPU, duration=duration)
    return _GPU(task, duration)


def _GPU(
    task: Callable[P, R],
    duration: int | timedelta | None,
) -> Callable[Concatenate[gr.Request, P], R]:

    if not Config.zero_gpu:
        # TODO: still prepend gr.Request for type consistency ?
        return task # type: ignore

    if task in decorated_cache:
        # TODO: Assert same duration ?
        return decorated_cache[task] # type: ignore

    if inspect.iscoroutinefunction(task):
        raise NotImplementedError

    if duration is None or isinstance(duration, timedelta):
        timedelta_duration = duration
    else:
        timedelta_duration = timedelta(seconds=duration)

    if inspect.isgeneratorfunction(task):
        decorated = generator_function_wrapper(task, timedelta_duration)
    else:
        decorated = regular_function_wrapper(task, timedelta_duration)

    client.startup_report()
    decorated_cache.update({
        task:      decorated,
        decorated: decorated,
    })

    return decorated # type: ignore
